import click 
from urllib.parse import urlparse
from .home import get_home
from .merge import merge
import tomlkit

@click.command()
@click.option("--ours", is_flag=True)
@click.option("--theirs", is_flag=True)
@click.option("--union", is_flag=True)
def pull(ours, theirs, union):
    home_path = get_home()
    
    with home_path.open(mode="r") as home_fd:
        home = tomlkit.parse(home_fd.read())
        for upstream in home.keys():
            if "origin" not in home[upstream]:
                click.echo(f"{upstream}", err=True)
                exit(1)
            dist = home_path.parent.joinpath(home[upstream].get("folder") or upstream)
            if not dist.exists():
                click.echo(f"{str(dist)} does not exist!", err=True)
                exit(1)
            merge(home[upstream]["origin"], dist, False, ours, theirs, union)
            


# SPDX-FileCopyrightText: 2022 Fiana Fortressia
#
# SPDX-License-Identifier: MIT OR APACHE-2.0

from typing import *
import click
class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None,  **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)

    def format_commands(self, ctx, formatter: click.HelpFormatter):
        super().get_usage(ctx)
        formatter.write_paragraph()

        main_commands = list()
        build_commands = list()
        
        for command in super().list_commands(ctx):
                main_commands.append(command)

        with formatter.section("The available commands are"):
            for command in main_commands:
                command: click.Command = super().get_command(ctx, command)
                if command.help:
                    desc = command.help.split("\n")[0]
                else:
                    desc = "No description available"
                formatter.write_text(f"{command.name} - {desc}")
      


# SPDX-FileCopyrightText: 2022 Fiana Fortressia
#
# SPDX-License-Identifier: MIT OR APACHE-2.0

import click 
from pathlib import Path 
import os
from typing import Optional 
def _get_home_from_cwd(parent: Optional[Path]):
    cwd = parent or Path.cwd()
    home = cwd.joinpath("flaura.toml")
    if home.exists():
        return home
    else:
        if cwd.parent == Path.home() or cwd.parent == cwd.root:
            click.echo("We couldn't find any flaura.toml file!! Please initiate them first by adding an upstream!!! See flaura upstream -h", err=True)
            exit(1)
        _get_home_from_cwd(cwd.parent)
def get_home():
    return Path(os.environ.get("FLAURA_HOME") or _get_home_from_cwd(None))


@click.command()
def init():
    """Initialize a flaura upstreams file in the current working directory!!"""
    Path.cwd().joinpath("flaura.toml").touch()


# SPDX-FileCopyrightText: 2022 Fiana Fortressia
#
# SPDX-License-Identifier: MIT OR APACHE-2.0

from pathlib import Path
import click 
from .home import get_home
import tomlkit
from os import path

@click.command()
@click.argument("origin")
@click.argument("path", required=False)
@click.option("--home", is_flag=True)
def add(origin, path, home):
    """Add an upstream! 
    
    """
    home_path = get_home()
    if home:
        path = home_path.parent()
    else: 
        if not path:
            print("Please specify a path!!")
            exit(1)
        path = Path(path)
    with home_path.open(mode="r") as home_fd:
        home = tomlkit.parse(home_fd.read())
        upstream = tomlkit.table()
        upstream["origin"] = origin
        home.append(str(path), upstream)

        with home_path.open(mode="w") as home_fd:
            home_fd.write(tomlkit.dumps(home))
    
@click.command()
@click.argument("name", type=str)
def remove(name):
    home_path = get_home()
    
    with home_path.open(mode="r") as home_fd:
        home = tomlkit.parse(home_fd.read())
        home.remove(name)
        with home_path.open(mode="w") as home_fd:
            home_fd.write(tomlkit.dumps(home))



@click.command("list")
def list_sources_cmd():
     list_sources()

def list_sources():
    home_path = get_home()
    with home_path.open(mode="r") as home_fd:
        home = tomlkit.parse(home_fd.read())
        for upstream in home.keys():
            click.secho(" " + upstream + " ", bg="green", fg="bright_white")
            click.echo(click.style("url: ", bold=True) + home[upstream]["origin"])
            click.echo(click.style("folder: ", bold=True) + (home[upstream].get("folder") or str(home_path.parent.joinpath(upstream))) + "\n")
            



# SPDX-FileCopyrightText: 2022 Fiana Fortressia
#
# SPDX-License-Identifier: MIT OR APACHE-2.0

from flaura.pull import pull
from .home import init
from .upstream import add, list_sources_cmd, remove
import click 
from .merge import merge, merge_cmd
from .group import OrderedGroup 

@click.group(cls=OrderedGroup)
def flaura():
    """Manage upstreams with ease!!
    Flaura is a tool for . It has replacements for Git submodules.
    """
    pass

flaura.add_command(merge_cmd)
flaura.add_command(init)
flaura.add_command(pull)
flaura.add_command(add)
flaura.add_command(remove)

flaura.add_command(list_sources_cmd)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flaura']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0', 'tomlkit>=0.11.5,<0.12.0']

entry_points = \
{'console_scripts': ['flaura = flaura:flaura']}

setup_kwargs = {
    'name': 'flaura',
    'version': '0.2.0',
    'description': 'Merge upstream with ease!',
    'long_description': "# Flaura\nMerge upstream with ease!\n\nFlaura implements all the merging behaviors of Git, and not only that, it's Git-ndependent, or standalone, which makes merging a non-Git folder, or just a certain folder in a Git repository, instead of the whole repository. \n\nIt allows hardforking to be like a charm!!\n\n## Project Status\nAt the moment, Flaura is still in early development. \n\nNo commands have been implemented except `flaura merge` and `flaura init`. \n\n`flaura merge` should be the basic functionality of Flaura for it to be useful, even though the UX is lacking!! You can see its documentation with `flaura merge --help`",
    'author': 'Fiana Fortressia',
    'author_email': 'fortressnordlys@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/waylovely-project/flaura.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2022 Fiana Fortressia
#
# SPDX-License-Identifier: MIT OR APACHE-2.0

from collections import namedtuple
import hashlib
import shutil 
from pathlib import Path
import subprocess
from urllib.parse import ParseResult, urlparse, urlunparse
from .home import get_home
import click 
import filecmp
@click.command("merge")
@click.argument("origin", required=False)
@click.argument("current_path", type=Path, required=False)
@click.option("--home", is_flag=True)
@click.option("--ours", is_flag=True)
@click.option("--theirs", is_flag=True)
@click.option("--union", is_flag=True)
def merge_cmd(origin, current_path, home, ours, theirs, union): 
    """Merge a URI source with a folder. At the moment only supports file:// URI.
    Uses the GNU RCS' merge command if available, but will use Git's merge-file command if Git is installed!!

    If there is a conflict, Git merge conflicts will appear in the files, and you can use Git integrations in your code editors to resolve them! 
    Pretty much this command runs like Git's behavior!

    origin: "The URI source to merge into current_path. At the moment. Example: file://../smithay/anvil"
    """
    merge(origin, current_path, home, ours, theirs, union)
    
def merge(origin, current_path: Path, home, ours, theirs, union):
    url = urlparse(origin)

    if url.scheme == "file":
        path = Path(url.netloc+"/"+url.path)

        merge_inner(path, current_path, home=home, ours=ours, theirs=theirs, union=union)
    if url.scheme.startswith("git+") or url.netloc == "github.com":
        url2 = ParseResult(url.scheme.removeprefix("git+"), url.netloc, url.path, url.params, url.query, url.fragment)
        dist = get_home().parent.joinpath(".flaura", "origins", 
        (url.scheme + "+" + url.netloc+"+"+url.path.split("/")[-2] + "/" +url.path.split("/")[-1] + "+" +  hashlib.md5(origin.encode("utf-8")).hexdigest() ) )
        if not dist.exists():
            print(dist)
            dist.mkdir(parents=True)
            subprocess.run(["git", "clone", urlunparse(url2), dist], cwd=dist)

        merge_inner(dist, current_path, home=home, ours=ours, theirs=theirs, union=union)

    else:
        click.echo(f"{url.scheme} for {origin} is not yet implemented. Sorry!! <3", err=True)
        exit(1)
    
        
        

def merge_inner(parent: Path, target_path: Path, **kwargs):
    if not target_path.exists():
        target_path.mkdir(parents=True)
    for path in parent.iterdir():
        target = target_path.joinpath(path.name)
        if path.is_dir():
           merge_inner(path, target) 
        else:
            if target.exists():
                if not filecmp.cmp(path, target):
                    try:
                        path.read_text()

                        suceed = True
                    except UnicodeDecodeError:
                        shutil.move(target, target.with_suffix(".older"))
                        shutil.copy(path, target)
                    
                    if suceed:
                        
                        if shutil.which("merge"):
                            command = ["merge"]
                        elif shutil.which("git"):
                            command = ["git", "merge-file"]
                        options = []

                        if kwargs.get("ours"):
                            options.push( "--ours" )
                        elif kwargs.get("theirs"):
                            options.push( "--theirs" ) 
                        elif kwargs.get("diff3"):
                            options.push( "--diff3" ) 
                        elif kwargs.get("zdiff3"):
                            options.push( "--zdiff3" )
                        elif kwargs.get("union"):
                            options.push( "--union" )
                            
                        subprocess.run([*command, target, path, target, 
                       ], check=True)
            else:
                shutil.copyfile(path, target)
        